from .device import Device, _name_attr, _inc,  GROUP
from . import keys as K
from .manager import Manager # see at the end of file
from .interface import UInt32, Int32, nodealiasproperty
from .tools import upload



#                      _              _   

#   ___ ___  _ __  ___| |_ __ _ _ __ | |_ 
#  / __/ _ \| '_ \/ __| __/ _` | '_ \| __|

# | (_| (_) | | | \__ \ || (_| | | | | |_ 

#  \___\___/|_| |_|___/\__\__,_|_| |_|\__|

# 


##### ###########
#
# SUBSTATE

@_name_attr

class SUBSTATE:
    
    NONE                  =   0
    
    NOTOP_NOTREADY        = 100
    
    NOTOP_INITIALISING    = 102
    
    NOTOP_READY_OFF       = 103
    
    NOTOP_READY_ON        = 104
    
    NOTOP_ERROR           = 199
    
    OP_DISABLING          = 205
    
    OP_OFF                = 206
    
    OP_SWITCHING_OFF      = 207
    
    OP_COOLING            = 208
    
    OP_ON                 = 209
    
    OP_SWITCHING_ON       = 210
    
    OP_WARMING            = 211
    
    OP_ERROR              = 299

    group = {
        NONE                   : GROUP.UNKNOWN,
        NOTOP_NOTREADY         : GROUP.NOK,
        NOTOP_READY_OFF        : GROUP.NOK,
        NOTOP_READY_ON         : GROUP.NOK,
        NOTOP_INITIALISING     : GROUP.BUZY,
        NOTOP_ERROR            : GROUP.ERROR, 
  
        OP_DISABLING            : GROUP.BUZY, 
        OP_SWITCHING_OFF        : GROUP.BUZY,
        OP_SWITCHING_ON         : GROUP.BUZY,

        OP_COOLING              : GROUP.BUZY,
        OP_WARMING              : GROUP.BUZY,
        OP_ON                   : GROUP.OK,
        OP_OFF                  : GROUP.OK,
        OP_ERROR                : GROUP.ERROR,    
    }


### #############
# ERROR
@_name_attr
class ERROR:
  
    OK					= _inc(0)
    HW_NOT_OP           = _inc()
    INIT_FAILURE        = _inc()		
    UNEXPECTED_OFF      = _inc()
    UNEXPECTED_ON       = _inc()
    FAULT_SIG           = _inc()
    MAXON               = _inc()
    STILL_COOLING       = _inc()
    TIMEOUT_DISABLE     = _inc()
    TIMEOUT_INIT        = _inc()
    TIMEOUT_OFF         = _inc()
    TIMEOUT_ON          = _inc()
    # Simulator errors
    SIM_NOT_INITIALISED	= 90
    SIM_NULL_POINTER    = 100	
    txt = {
        OK:					 'OK',
	    HW_NOT_OP:			 'ERROR: TwinCAT not in OP state or CouplerState not mapped.',
	    INIT_FAILURE:		 'ERROR: INIT command aborted due to STOP or RESET.',
	    UNEXPECTED_OFF:		 'ERROR: Lamp unexpectedly switched OFF.',
	    UNEXPECTED_ON:		 'ERROR: Lamp unexpectedly switched ON.',
	    FAULT_SIG:			 'ERROR: Fault signal active.',
	    MAXON:				 'ERROR: Lamp maximum ON time exceeded.',
	    STILL_COOLING:		 'ERROR: ON command not allowed while cooling.',
	    TIMEOUT_DISABLE:	 'ERROR: Disable timed out.',
	    TIMEOUT_INIT:		 'ERROR: Init timed out.',
	    TIMEOUT_OFF:		 'ERROR: Switching OFF timed out.',
	    TIMEOUT_ON:			 'ERROR: Switching ON timed out.',
	    SIM_NOT_INITIALISED: 'ERROR: Lamp simulator not initialised.',
	    SIM_NULL_POINTER:	 'ERROR: NULL pointer to Lamp.',
    }



### #############
#
# RPC error

@_name_attr

class RPC_ERROR:
    
    OK                 =  0 
    
    NOT_OP             = -1
    
    NOT_NOTOP_READY    = -2
    
    NOT_NOTOP_NOTREADY = -3
    
    SWITCHING_ON       = -4
    
    SWITCHING_OFF      = -5
    
    COOLING            = -6
    
    LOCAL              = -7
    txt = {
        OK:						 'OK',
	    NOT_OP:					 'Cannot control lamp. Not in OP state.',
	    NOT_NOTOP_READY:		 'Call failed. Not in NOTOP_READY.',
	    NOT_NOTOP_NOTREADY:		 'Call failed. Not in NOTOP_NOTREADY/ERROR.',
	    SWITCHING_ON:			 'Lamp OFF failed. Still switching ON.',
	    SWITCHING_OFF:			 'Lamp ON failed. Still switching OFF.',
	    COOLING:				 'Lamp ON failed. Still cooling down.',
	    LOCAL:					 'RPC calls not allowed in Local mode',
    }

#  _       _             __                

# (_)_ __ | |_ ___ _ __ / _| __ _  ___ ___ 
# | | '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \

# | | | | | ||  __/ |  |  _| (_| | (_|  __/

# |_|_| |_|\__\___|_|  |_|  \__,_|\___\___|



class LampStatInterface(Device.StatInterface):
    ERROR = ERROR
    SUBSTATE = SUBSTATE

    @nodealiasproperty("is_off", ["substate"])
    def is_off(self, substate):
        return substate == Lamp.SUBSTATE.OP_OFF
    
    @nodealiasproperty("is_on", ["substate"])
    def is_on(self, substate):
        return substate == Lamp.SUBSTATE.OP_ON
    
class LampCfgInterface(Device.CfgInterface):
    
    # we can define the type to parse value directly on the class by annotation
    
    analog_threshold : Int32
    analog_range : UInt32
    cooldown: UInt32
    maxon : UInt32 
    warmup : UInt32
    timeout : UInt32
    
    # redefine the Method interface to include the proper description of the RpcError
class LampRpcNode(Device.RpcInterface.RpcNode):
    RPC_ERROR = RPC_ERROR


class LampRpcInterface(Device.RpcInterface):
    
    RpcNode = LampRpcNode
    
    RPC_ERROR = RPC_ERROR
    
    ##
    # the type of rpcMethod argument can be defined by annotation
    
    # All args types must be defined in a tuple
    
    rpcSwitchOn : (float, UInt32)
    
    



#      _            _          
#   __| | _____   _(_) ___ ___ 
#  / _` |/ _ \ \ / / |/ __/ _ \
# | (_| |  __/\ V /| | (_|  __/
#  \__,_|\___| \_/ |_|\___\___|
#

class Lamp(Device):
    
    SUBSTATE = SUBSTATE
    
    ERROR = ERROR
      
    RpcInterface = LampRpcInterface
    
    StatInterface = LampStatInterface
    
    CfgInterface = LampCfgInterface
    
    def switch_on(self, intensity, time_limit):
 
        """ switch on the lamp 
        
        Args:
            intensity (float): in % 
            time_limit (float): number of second the lamp will stay on
        """       
        # intensity - float, onTimeLimit - integer
        
        self.rpc.rpcSwitchOn.rcall(intensity, onTimeLimit)

        
    def switch_off(self):
        """ switch off the lamp """        
        self.rpc.rpcSwitchOff.rcall()
        
            
     
    


Manager.record_new_device_type('Lamp', Lamp)

Manager.record_cfg_template('Lamp', """{name}:
  type: Lamp
  interface: Softing
  identifier: PLC1                             # OPCUA Object Identifier
  prefix: MAIN.Lamp1                           #  TO BE UPDATED  OPCUA attribute prefix
  simulated: false
  ignored: false
  address: {address}      # TO BE UPDATED 
  simaddr: opc.tcp://134.171.12.182:4840      # Simulation address
  mapfile: "{cfg_dir}/mapLamp.yml"    #  TO BE UPDATED
  fits_prefix: "LAMP1"
  ctrl_config:
    low_fault:        false                     # If T, signal is active low
    low_on:           false                     # If T, signal is active low
    low_switch:       false                     # If T, signal is active low
    initial_state:    false
    timeout:         2000
""")


import yaml
import os

RESPATH = 'RESPATH'

def read_map(file):
    with open(file) as f:
        return yaml.load(f.read(), Loader=yaml.CLoader)

def read_config(file):    
    with open(file) as f:
        return yaml.load(f.read(), Loader=yaml.CLoader)

def load_config(file_name):
    return read_config(find_config(file_name))

def load_extra_of(file_name):
    
    root, ext = os.path.splitext(find_config(file_name))
    
    
    file_name = os.path.join(root+"_extra"+ext)
    if not os.path.exists(file_name):
        return None
    return read_config(file_name)

def load_map(file_name):
    return read_map(find_config(file_name))

def find_config(file_name):
    path_list = os.environ.get(RESPATH, '.').split(':')
    for directory in path_list[::-1]:
        path = os.path.join(directory, file_name)
        if os.path.exists(path):
            return  path
    raise ValueError('coud not find config file %r in any of %s'%(file_name, path_list))


    


from . import keys as K
from .device import Device, kjoin, fjoin, GROUP
from . import io
from .interface import NodeAlias, Interface, nodealiasproperty, ksplit, fsplit, fjoin
import os

warning = print # for now will see later

def setitem(obj,k,v):
    """ setitem(obj, k, v) -> obj[k] = v"""
    obj[k] = v

def open_manager(file_name, key='', extra_file=None):
    """ Create a :class:`Manager` object from its yaml configuration file 
    
    Args:
        file_name (str) : This is the relative path trought the yml file defining the manager. 
                The file shall be present inside one of the directories defined by the 
                $RESPATH environmnet variable.
        key (str, optional):  The key of the manager (which will be the prefix of all devices)
               If not given key will be the 'server_id' keyword has defined in the configuration file
        extra_file (str): This is the path to an extra yml file where some specific pydevmgr configuration 
               are writen. If not given an eventual FILE_NAME_extra.yml file will be load. 
               FILE_NAME_extra.yml must be on the same directory than FILE_NAME.yml
               So far in the _extra file is defined GUI configurations.
                 
    """
    return Manager.from_config(file_name, key=key, extra_file=extra_file)

def open_device(file_name, cfg_name, key=None, fits_prefix=''):
    """ read a device configuration file and return a :class:`Device` like object to handle it 
    
    Args:
        file_name (str): relative path to a configuration file 
                  The path is relative to one of the directory defined in the
                  $RESPATH environmnet variable
        cfg_name (str): The device name in the config file. 
                
            For instance in config file like
                
                :: 
                     
                    motor1:
                      type: Motor
                      interface: Softing
                      identifier: PLC1
                      etc... 
            
            'motor1' is the `cfg_name`
        
        key (str, optional): The device key. If not given cfg_name is taken
        fits_prefix (str, optional): The fits prefix string for this device
    """
    if key is None:
        key = cfg_name
    
    allconfig = io.load_config(file_name)
    
    try:
        config = allconfig[cfg_name]
    except KeyError:
        raise ValueError("Device %r does not exists on device definition file %r"%(cfg_name, file_name))
    try:
        mapfile = config[K.MAPFILE]
    except KeyError:
        raise ValueError("Could not find mapfile in device definition %r at %r"%(cfg_name, file_name))
    map_d = io.load_map(mapfile)
    try:
        dtype = config[K.TYPE]
    except KeyError:
        raise ValueError("config file does not define device 'type'")
    try:
        map = map_d[dtype]
    except KeyError:
        raise ValueError("The associated map file does not contain type %r"%dtype)
    try:
        cl = Manager.get_device_type(dtype)
    except ValueError:
        warning("The type %r is not implemented, %r will have a generic Device class"%(dtype, cfg_name))
        cl = Manager.Device
        
    return cl( key, config, map, fits_prefix=fits_prefix, config_file=file_name)  


## #######################################################
#
#   Some NodeAlias for the manager 
#   These are created in the stat attribute 
#

class SubstateNodeAlias(NodeAlias):
    SUBSTATE = Device.SUBSTATE
    def fget(self, *substates):
        if not substates: return self.SUBSTATE.UNKNOWN
        first = substates[0]
        if all( s==first for s in substates):
            return first 
        return self.SUBSTATE.UNKNOWN

class StateNodeAlias(NodeAlias):
    STATE = Device.STATE
    def fget(self, *states):
        if all( s==self.STATE.OP for s in states):
            return self.STATE.OP
        return self.STATE.NOTOP
        
class InitialisedNodeAlias(NodeAlias):    
    def fget(self, *init_flags):
        return all(f for f in init_flags)



##
# The stat manager for stat interface will be build of NodeAliases
#  !!!!!!!!!!!  
#      ManagerStatInterface must be initialised by Manager.stat property
#       Manager.stat should fill the NodeAlias(es) after devices list has been built
#  !!!!!!!!!!
class ManagerStatInterface(Interface):
    
    STATE = Device.STATE
    @nodealiasproperty("state_txt", ["state"])
    def state_txt(self, state):
        return self.STATE.txt.get(state, "not registered state")
    
    @nodealiasproperty("state_group", ["state"])
    def state_group(self, state):
        return self.STATE.group.get(state, GROUP.UNKNOWN)
    
        
class Manager:
    """ Manager object handling several devices 
    
    Most likely the Manager will be initialized by :func:`Manager.from_config` or its alias :func:`open_manager`
    
    
    Args:
        key (str): the key (prefix of all devices) of the manager
                If None key is the 'server_id' defined inside the config dictionary
        config (dic): The config dictionary (as red from yml config files by :func:`Manager.from_config`)
        fits_prefix (str, optional): optional prefix for the Fits keyword (if ever used)
    
    """
    # yes this dictionary is intended to be set on the class
    __device_types__ = {}
    __device_templates__ = {}
    Device = Device # default device class
    
    StatInterface = ManagerStatInterface    
    _stat = None

    def __init__(self, key, config, fits_prefix='', config_file='', extra=None):
        
        server_id = config[K.SERVER_ID]
        key = server_id if key is None else key
        prefix, name = ksplit(key)
        
        server_config = config[server_id]
        
        device_names = server_config[K.DEVICES]
        name = server_id if name is None else name
        
        fits_key = fjoin(fits_prefix, server_config.get(K.FITS_PREFIX, ''))  
        
        devices = {}
        
        self._key = key 
        
        self._fits_key = fits_key
        self._config_file = config_file # for future use 
        
        for device_name in device_names:
            try:
                dev_info = config[device_name]
            except KeyError:
                raise ValueError('device %r is not defined in the configuration file'%device_name)
            dev_type = dev_info[K.TYPE]
            cfg_gile = dev_info[K.CFGFILE]
            try:
                cl = self.get_device_type(dev_type)
            except ValueError:
                warning("The type %r is not implemented, %r will have a generic Device class"%(dev_type, device_name))
                cl = self.Device
                
            devices[device_name] = cl.from_config(cfg_gile, device_name, key=kjoin(self.key, device_name), fits_prefix=self.fits_key)
            
        self._devices = devices
        
        self._extra = {} if extra is None else extra

    def __getattr__(self,attr):
        try:
            return self.__dict__[attr]
        except KeyError:
            try:
                return self.get_device(attr)
            except ValueError:
                raise AttributeError("%r"%attr)
    
    
    @property
    def key(self):
        return self._key
    
    @property
    def prefix(self):
        return ksplit(self._key)[0]

    
    @property
    def name(self):
        return ksplit(self._key)[1]
        
    @property
    def fits_key(self):
        return self._fits_key
    
    @property
    def fits_name(self):
        return fsplit(self._fits_key)[1]

    @property
    def fits_prefix(self):
        return fsplit(self._fits_key)[0]
    
    @property
    def extra(self):
        return self._extra
    
    @property
    def stat(self):
        if self._stat is None:
            # TODO add the parser functions dictionay
            self._stat = self.StatInterface(self.key, None, {}, '')
            
            self._stat.state = StateNodeAlias(kjoin(self.key,"state"), [d.stat.state for d in self.devices()])
            
        return self._stat
    
    
    @classmethod
    def from_config(cl, file_name, key=None, extra_file=None):
        """ Create a :class:`Manager` object from its yaml configuration file 
        
        Args:
            file_name (str): This is the relative path trought the yml file defining the manager. 
                    The file shall be present inside one of the directories defined by the 
                    $RESPATH environmnet variable.
            key (str, optional):  The key of the manager (which will be the prefix of all devices)
                   If not given key will be the 'server_id' keyword has defined in the configuration file
            extra_file (str): This is the path to an extra yml file where some specific pydevmgr configuration 
                   are writen. If not given an eventual FILE_NAME_extra.yml file will be load. 
                   FILE_NAME_extra.yml must be on the same directory than FILE_NAME.yml
                   So far in the _extra file is defined GUI configurations.
                     
        """
        config = io.load_config(file_name)
        if extra_file:
            extra = io.load_config(extra_file)
        else:
            extra = io.load_extra_of(file_name)
        
        return cl(key, config, config_file=file_name, extra=extra)
        
    @classmethod
    def record_new_device_type(cl, type_name, dev_cls):
        """ record a new device to the Manager class
        WARNING: the recorded or overwriten type will have effect on all new instance
                  of Manager
        """
        if not hasattr(dev_cls, 'from_config'):
            raise ValueError('class must have the `from_config` method')
        cl.__device_types__[type_name] = dev_cls
        
    
    @classmethod
    def record_cfg_template(cl, name, template):
        cl.__device_templates__[name] = template    
        #if not hasattr(cl, type_name):
        #    setattr(cl, type_name, dev_cls) # this can have some side effect

    @classmethod
    def get_device_type(cl, type_name):
        try:
            return cl.__device_types__[type_name]
        except KeyError:
            raise ValueError('Unknown device type %r '%type_name)
    
    def connect_all(self):
        """ Connect all the opc-ua client of the devices """
        for device in self._devices.values():
            device.connect()

    def disconnect_all(self):
        """ disconnect all the opc-ua client of the devices """
        for device in self._devices.values():
            device.disconnect()
            
    def get_device(self, name):
        """ get device matching the name Raise ValueError if not found """
        try:
            return self._devices[name]
        except KeyError:
            raise ValueError('Unknown device %r'%name)

    def devices(self):
        """ return the list of children :class:`Device` like object """
        return list(self._devices.values())

    def device_names(self):
        """ return a list of child device names """
        return list(self._devices.keys())

    ## These RPC should be on all devices
    def init_all(self):
        """ Init all child devices """
        for device in self.devices():
            device.init()

    def enable_all(self):
        """ Enable all child devices """
        for device in self.devices():
            device.enable()

    def disable_all(self):
        """ Disable all child devices """
        for device in self.devices():
            device.disable()

    def reset_all(self):
        """ Reset all child devices """
        for device in self.devices():
            device.reset()
    
    def configure_all(self):
        """ Configure all child devices """
        for device in self.devices():
            device.configure()


Manager.record_cfg_template('Manager', """
server_id           : '{name}'

{name}:
    req_endpoint    : "zpb.rr://127.0.0.1:12082/"                             # Not used in pydevmgr 
    pub_endpoint    : "zpb.ps://127.0.0.1:12345/"                             #
    db_endpoint     : "127.0.0.1:6379"                                        #
    db_timeout      : 2                                                       #
    scxml           : "{cfg_dir}/sm.xml"                              #
    dictionaries    : ['dit/stddid/primary.did', 'fcf/devmgr/server/fcf.did'] #
    fits_prefix     : "FCS1"
    #devices         : ['shutter1', 'lamp1', 'motor1', 'adc1', 'drot1', 'sensor1', 'ccc1', 'piezo1', 'actuator1']
    devices         : ['shutter1', 'lamp1',  'piezo1', 'motor1', 'motor2', 'adc1', 'drot1']
    cmdtout         : 60000

shutter1:
  type: Shutter
  cfgfile: "{cfg_dir}/shutter1.yml"

lamp1:
  type: Lamp
  cfgfile: "{cfg_dir}/lamp1.yml"

motor1:
  type: Motor
  cfgfile: "{cfg_dir}/motor1.yml"

motor2:
  type: Motor
  cfgfile: "{cfg_dir}/motor2.yml"
  
sensor1:
  type: Sensor
  cfgfile: "{cfg_dir}/sensor1.yml"

drot1:
  type: Drot
  cfgfile: "{cfg_dir}/drot1.yml"

adc1:
  type: Adc
  cfgfile: "{cfg_dir}/adc1.yml"

piezo1:
  type: Piezo
  cfgfile: "{cfg_dir}/piezo1.yml"

""")    


from .device import Device, _name_attr, _inc, GROUP
from . import keys as K
from .manager import Manager # see at the end of file
from .interface import nodealiasproperty, UInt32
from .tools import upload

#                      _              _   
#   ___ ___  _ __  ___| |_ __ _ _ __ | |_ 
#  / __/ _ \| '_ \/ __| __/ _` | '_ \| __|
# | (_| (_) | | | \__ \ || (_| | | | | |_ 
#  \___\___/|_| |_|___/\__\__,_|_| |_|\__|
# 


@_name_attr
class SUBSTATE:
    OFF = 0
    NOT_OP = 1
    OP = 2
    NOT_OP_NOT_READY = 100
    NOT_OP_INITIALIZING = 102
    NOT_OP_READY_CLOSED = 105
    NOT_OP_READY_OPEN = 106
    NOT_OP_FAILURE = 199
    OP_DISABLING = 205
    OP_CLOSED = 212
    OP_CLOSING = 213
    OP_OPEN = 214
    OP_OPENING = 215
    OP_FAILURE = 299
    group = {
        OFF                   : GROUP.UNKNOWN,
        OP                    : GROUP.OK,
        NOT_OP_NOT_READY      : GROUP.NOK,
        NOT_OP_READY_CLOSED   : GROUP.NOK,
        NOT_OP_READY_OPEN     : GROUP.NOK,
        NOT_OP_INITIALIZING   : GROUP.BUZY,
        NOT_OP_FAILURE        : GROUP.ERROR,
        OP_DISABLING          : GROUP.BUZY, 
        OP_CLOSING            : GROUP.BUZY, 
        OP_OPENING            : GROUP.BUZY,
        OP_FAILURE            : GROUP.ERROR,
        OP_CLOSED             : GROUP.OK, 
        OP_OPEN               : GROUP.OK,    
    }
    


@_name_attr
class ERROR:
    OK				      = _inc(0)
    HW_NOT_OP             = _inc()			
    INIT_FAILURE          = _inc()	
    UNEXPECTED_CLOSED     = _inc()	
    UNEXPECTED_NONE       = _inc()	
    UNEXPECTED_OPENED     = _inc()	
    FAULT_SIG             = _inc()	
    BOTH_SIG_ACTIVE       = _inc()
    TIMEOUT_ENABLE        = _inc()
    TIMEOUT_DISABLE       = _inc()
    TIMEOUT_INIT          = _inc()
    TIMEOUT_CLOSE         = _inc()
    TIMEOUT_OPEN          = _inc()
    
    SIM_NOT_INITIALISED		= 90
    SIM_NULL_POINTER		= 100	# Simulator error
    txt = {
        OK:				   'OK',
        HW_NOT_OP:		   'ERROR: TwinCAT not in OP state or CouplerState not mapped.',
        INIT_FAILURE:	   'ERROR: INIT command aborted due to STOP or RESET.',
        UNEXPECTED_CLOSED: 'ERROR: Shutter unexpectedly closed.',
        UNEXPECTED_NONE:   'ERROR: Unexpectedly no OPEN or CLOSED signal active.',
        UNEXPECTED_OPENED: 'ERROR: Shutter unexpectedly opened.',
        FAULT_SIG:		   'ERROR: Fault signal active.',
        BOTH_SIG_ACTIVE:   'ERROR: Both OPEN and CLOSED signals active.',
        TIMEOUT_ENABLE:	   'ERROR: ENABLE timed out.',
        TIMEOUT_DISABLE:   'ERROR: DISABLE timed out.',
        TIMEOUT_INIT:	   'ERROR: INIT timed out.',
        TIMEOUT_CLOSE:	   'ERROR: CLOSE timed out.',
        TIMEOUT_OPEN:		  'ERROR: OPEN timed out.',
        SIM_NOT_INITIALISED:  'ERROR: Shutter simulator not initialised.',
        SIM_NULL_POINTER:	  'ERROR: NULL pointer to Shutter.',
    }
@_name_attr
class RPC_ERROR:
    OK =  0
    NOT_OP				= -1			
    NOT_NOTOP_READY		= -2			
    NOT_NOTOP_NOTREADY	= -3	
    STILL_OPENING = -4
    STILL_CLOSING = -5
    LOCAL = -6
    txt = { # copy past on MgetRpcErrorTxt
        OK:					 'OK',
	    NOT_OP:				 'Cannot control shutter. Not in OP state.',
	    NOT_NOTOP_READY:	 'Call failed. Not in NOTOP_READY.',
	    NOT_NOTOP_NOTREADY:	 'Call failed. Not in NOTOP_NOTREADY/ERROR.',
	    STILL_OPENING:		 'Not allowed to close the shutter while opening.',
	    STILL_CLOSING:		 'Not allowed to open the shutter while closing.',
	    LOCAL:				 'RPC calls not allowed in Local mode.',
    }



#  _       _             __                
# (_)_ __ | |_ ___ _ __ / _| __ _  ___ ___ 
# | | '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \
# | | | | | ||  __/ |  |  _| (_| | (_|  __/
# |_|_| |_|\__\___|_|  |_|  \__,_|\___\___|

class ShutterStatInterface(Device.StatInterface):
    
    
    
    ERROR = ERROR
    SUBSTATE = SUBSTATE
        
    @nodealiasproperty("is_open", ["substate"])
    def is_open(self, substate):
        return substate == self.SUBSTATE.OP_OPEN
    
    @nodealiasproperty("is_closed", ["substate"])
    def is_closed(self, substate):
        return substate == self.SUBSTATE.OP_CLOSED
    
class ShutterCfgInterface(Device.CfgInterface):
    # we can define the type to parse value directly on the class by annotation
    timeout: UInt32  

# redefine the Method interface to include the proper description of the RpcError
class ShutterRpcNode(Device.RpcInterface.RpcNode):
    RPC_ERROR = RPC_ERROR

class ShutterRpcInterface(Device.RpcInterface):
    RpcNode = ShutterRpcNode
    RPC_ERROR = RPC_ERROR
    

#      _            _          
#   __| | _____   _(_) ___ ___ 
#  / _` |/ _ \ \ / / |/ __/ _ \
# | (_| |  __/\ V /| | (_|  __/
#  \__,_|\___| \_/ |_|\___\___|
#
class Shutter(Device):
    SUBSTATE = SUBSTATE
    RpcInterface = ShutterRpcInterface
    StatInterface = ShutterStatInterface
    CfgInterface = ShutterCfgInterface

    def open(self):
        """ open the shutter """
        self.rpc.rpcOpen.rcall()    
    
    def close(self):
        """ close the shutter """
        self.rpc.rpcClose.rcall()
        
    def stop(self):
        """ stop any motion """
        self.rpc.rpcStop.rcall()
    
Manager.record_new_device_type('Shutter', Shutter)

Manager.record_cfg_template('Shutter', """{name}:
  type: Shutter
  interface: Softing
  identifier: PLC1                             # OPCUA Object Identifier
  prefix: MAIN.Shutter1                       # OPCUA attribute prefix
  simulated: false
  ignored: false
  address: {address}        # TO BE UPDATED  PLC address (OPCUA Server)
  simaddr: opc.tcp://134.171.2.213:7576        # Simulation address
  mapfile: "{cfg_dir}/mapShutter.yml"  # TO BE UPDATED 
  fits_prefix: "SHUT1"
  ctrl_config:
    low_closed:     false
    low_fault:      false                              # If T, signal is active low
    low_open:       false                              # If T, signal is active low
    low_switch:     false                              # If T, signal is active low
    ignore_closed:  false                              # If T, ignore the signal
    ignore_fault:   false                              # If T, ignore the signal
    ignore_open:    false                              # If T, ignore the signal
    initial_state:  false
    timeout:       2000
""")




from PyQt5 import QtWidgets, uic
from .io import find_ui
from .generic import record_widget, method_caller, method_switcher, RpcError
from pydevmgr import Drot, Prefixed, download
from .basic import BaseCtrl, BaseLine

class MOVE_MODE:
    ABSOLUTE = 0
    RELATIVE = 1
    VELOCITY = 2
    txt = {
       ABSOLUTE : "ABSOLUTE", 
       RELATIVE : "RELATIVE", 
       VELOCITY : "VELOCITY"
    }
    count = 3

class TRACK_MODE:
    SKY = 2
    ELEV = 3
    USER = 4
    txt = {
        SKY  : 'SKY', 
        ELEV : 'ELEV',
        USER : 'USER'
    }
    ordered = [SKY, ELEV, USER]
    count = 3


class DrotCtrl(BaseCtrl):
    
    def initUI(self):
        uic.loadUi(find_ui('drot_ctrl_frame.ui'), self)
        
    def update_device(self, data, drot):
        super(DrotCtrl, self).update_device(data, drot)
        #self.input_pos_target1.update()
        #self.input_pos_target2.update()
        #self.input_velocity.update() 
        
        if self.move_mode.currentIndex()==MOVE_MODE.VELOCITY: # VELOCITY
            self.input_pos_target.setEnabled(False)
        else:
            self.input_pos_target.setEnabled(True)
        
        
        self.pos_actual.setText( "{:.3f}".format(data['pos_actual']) )
        self.pos_error.setText( "{:.3E}".format(data['pos_error']) )
        
        self.pos_target.setText( "{:.3f}".format(data['pos_target']) )
        self.vel_actual.setText( "{:.3f}".format(data['vel_actual']) )
        
        self.angle_on_sky.setText( "{:.2f}".format(data['angle_on_sky']) )
        
        self.track_mode_txt.setText( "{}".format(data['track_mode_txt']) )
        
    def link_device(self, downloader, drot, altname=None):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        drot (:class:`pydevmgr.Drot`):  Drot device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        super(DrotCtrl, self).link_device(downloader, drot, altname)
        token = self.token # comming from super 
        try:
            self.move.clicked.disconnect()
            self.start_track.disconnect()
            self.stop_track.disconnect()
        except TypeError:
            pass
        
        
        # init some field to what the state is 
        
        if drot.is_connected():
            v,p,a = download([drot.cfg.velocity, drot.stat.pos_actual, drot.stat.angle_on_sky])
            self.input_velocity.setText("%.3f"%v)
            self.input_pos_target.setText("%.3f"%p) 
            self.input_angle.setText("%.3f"%a)
  
        else:
            # put some dummy values
            self.input_velocity.setText("1.00")
            self.input_pos_target.setText("0.00")   
            self.input_angle.setText("0.0")
        
        self.track_mode.clear()
        self.track_mode.addItems(TRACK_MODE.txt[M] for M in TRACK_MODE.ordered)    
        
        self.move_mode.clear()
        self.move_mode.addItems(MOVE_MODE.txt[i] for i in range(MOVE_MODE.count))
        
        # add necessary nodes to the downloader state, substate, status, error_code 
        # are handled  by BaseCtrl.link_device
        
        downloader.add_node(token, 
                         drot.stat.track_mode_txt, 
                         drot.stat.pos_actual, drot.stat.pos_error, 
                         drot.stat.pos_target, drot.stat.vel_actual,# stat.vel_target,
                         drot.stat.angle_on_sky
                        )
        
        
        def move(idx,  pos, vel):
            if idx == MOVE_MODE.ABSOLUTE:
                drot.move_abs( pos, vel)
            elif idx == MOVE_MODE.RELATIVE:
                drot.move_rel( pos, vel)
            elif idx == MOVE_MODE.VELOCITY:
                drot.move_vel( vel)
            
        # Now link the buttons to an action
        move = method_caller(move, 
                            [self.move_mode.currentIndex, 
                             self.input_pos_target.text, 
                             self.input_velocity.text
                             ],
                             self.feedback
                            )
        self.move.clicked.connect(move)
        
        stop = method_caller(drot.stop, [], self.feedback)
        self.stop.clicked.connect(stop)
        
        move_angle = method_caller(drot.move_angle, [self.input_angle.text], self.feedback)
        self.move_angle.clicked.connect(move_angle)
        
        def track_mode():
            return TRACK_MODE.ordered[self.track_mode.currentIndex()]
        
        start_track = method_caller(drot.start_track, 
                                   [track_mode, self.input_angle.text], 
                                   self.feedback)
        self.start_track.clicked.connect(start_track)
        
        stop_track = method_caller(drot.stop_track, [], self.feedback)
        self.stop_track.clicked.connect(stop_track)
        
        

class DrotLine(BaseLine):    
    def initUI(self):
        uic.loadUi(find_ui('drot_line_frame.ui'), self)
        
    def update_device(self, data, drot):
        super(DrotLine, self).update_device(data, drot)
        
        
        self.pos_actual.setText( "{:.3f}".format(data['pos_actual']) )
        self.angle_on_sky.setText( "{:.2f}".format(data['angle_on_sky']) )

        self.track_mode_txt.setText( "{}".format(data['track_mode_txt']) )
    
    
    def link_device(self, downloader, drot, altname=None, more_methods=[]):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        drot (:class:`pydevmgr.Drot`):  Drot device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        methods = [
             "---", # is a separator and is not counted as item 
            ("MOVE ANGLE", drot.move_angle, [self.input_angle.text]),
            ("TRK ELEV", drot.start_track,[lambda : TRACK_MODE.ELEV, self.input_angle.text]),
            ("TRK SKY", drot.start_track, [lambda : TRACK_MODE.SKY , self.input_angle.text]),
            ("TRK USER", drot.start_track,[lambda : TRACK_MODE.USER, self.input_angle.text]), 
            ("STOP TRK", drot.stop_track, [] ), 
        ]+ list(more_methods)
                        
        super(DrotLine, self).link_device(downloader, drot, altname, more_methods=methods)
        token = self.token # comming from super 
        
        self.input_angle.setText("{:.2f}".format(downloader.data.get(drot.stat.angle_on_sky.key, 0.0)))        
        
        downloader.add_node(token, 
                         drot.stat.pos_actual,drot.stat.angle_on_sky, 
                         drot.stat.track_mode_txt
                        )
        

record_widget("ctrl", "Drot", DrotCtrl)
record_widget("line", "Drot", DrotLine)




from PyQt5 import QtWidgets, uic
from .io import find_ui
from .generic import record_widget, method_caller, method_switcher, RpcError
from pydevmgr import Lamp, Prefixed
from .basic import BaseCtrl, BaseLine

class LampCtrl(BaseCtrl):
    
    def initUI(self):
        uic.loadUi(find_ui('lamp_ctrl_frame.ui'), self)
        
    def update_device(self, data, lamp):
        super(LampCtrl, self).update_device(data, lamp)
        self.input_intensity.update()
        self.input_time.update() 
        
        self.time_left.setText( "{:.0f}".format(data['time_left']) )
        self.intensity.setText( "{:.1f}".format(data['intensity']) )
        
    def link_device(self, downloader, lamp, altname=None):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        lamp (:class:`pydevmgr.Lamp`):  Lamp device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        super(LampCtrl, self).link_device(downloader, lamp, altname)
        downloader, token = self.connection # comming from super 
        try:
            self.on.clicked.disconnect()
            self.off.clicked.disconnect()
        except TypeError:
            pass
        
        
        # put some start values, a pity that there is no way to get the last 
        # entered value
        self.input_intensity.setText("1")
        self.input_time.setText("10")            
        
        # add necessary nodes to the downloader state, substate, status, error_code 
        # are handled  by BaseCtrl.link_device
        stat = lamp.stat
        downloader.add_node(token, 
                         stat.time_left, stat.intensity
                        
                        )
        
        # Now link the buttons to an action
        on = method_caller(lamp.switch_on, 
                                 [self.input_intensity.text, self.input_time.text],
                                 self.feedback
                                 )
        
        self.on.clicked.connect(on)
        
        
        off = method_caller(lamp.switch_off, [], self.feedback)
        self.off.clicked.connect(off)


class LampLine(BaseLine):    
    def initUI(self):
        uic.loadUi(find_ui('lamp_line_frame.ui'), self)
        self.check.hide() # hide check box by default it is show by the manager panel

    def update_device(self, data, lamp):
        super(LampLine, self).update_device(data, lamp)
        self.input_intensity.update()
        self.input_time.update() 
        
        self.time_left.setText( "{:.0f}".format(data['time_left']) )
        
    def link_device(self, downloader, lamp, altname=None, more_methods=[]):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        lamp (:class:`pydevmgr.Lamp`):  Lamp device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        more_methods = [
                        ("ON",  lamp.switch_on, [self.input_intensity.text, self.input_time.text]),
                        ("OFF", lamp.switch_off, [])
                        ]+list(more_methods)
        super(LampLine, self).link_device(downloader, lamp, altname, 
                                          more_methods=more_methods)
        token = self.token # comming from super 
        
        # put some start values, a pity that there is no way to get the last 
        # entered value
        self.input_intensity.setText("1")
        self.input_time.setText("10")            
        
        # add necessary nodes to the downloader state, substate, status, error_code 
        # are handled  by BaseCtrl.link_device
        stat = lamp.stat
        downloader.add_node(token, 
                         stat.time_left, stat.intensity
                        
                        )


record_widget("ctrl", "Lamp", LampCtrl)
record_widget("line", "Lamp", LampLine)


#!/usr/bin/env python

import os
import sys

#from distutils.core import setup # dependency does not works
from setuptools import setup

name = "pydevmgr"
names = ["pydevmgr", "pydevmgr_qt"] # if None build only the package defined by name

version='0.1.2' # https://www.python.org/dev/peps/pep-0440/
author='Sylvain Guieu'
author_email='sylvain.guieu@univ-grenoble-alpes.fr'
install_requires = ['opcua', 'pyyaml']
# in setuptools the egg fragment must have the version in it
# unlike with the pip install git+https://gricad-gitlab.univ-grenoble-alpes.fr/guieus/instru#egg=instru
#dependency_links=['git+https://gricad-gitlab.univ-grenoble-alpes.fr/guieus/instru#egg=instrutools-0.1.dev0']
dependency_links=[]
# any file inside the data relative directory name/data_dir will be include as data file
data_directories = [(os.path.join('pydevmgr_qt','uis'), '.ui', 'pydevmgr_qt_uis')]#[(os.path.join(name,'resources'),'', 'pydevmgr_resources')]

#(os.path.join(name,'guis'),'.ui')]

script_directories = ["pydevmgr/scripts", "pydevmgr_qt/scripts"]
license='Creative Commons Attribution-Noncommercial-Share Alike license'




# Python 3.0 or later needed
if sys.version_info < (3, 0, 0, 'final', 0):
    raise SystemExit('Python 3.0 or later is required!')


## ######################################################
##
##  Try to make this part bellow stand alone so I can copy/past
##   to other projects
##
## ######################################################


rootdir = os.path.abspath(os.path.dirname(__file__))



# Build a list of all project modules
packages = []
if names:
    for _nme in names:
        for dirname, dirnames, filenames in os.walk(_nme):
            if '__init__.py' in filenames:
                packages.append(dirname.replace('/', '.'))
else:
    for dirname, dirnames, filenames in os.walk(name):
        if '__init__.py' in filenames:
            packages.append(dirname.replace('/', '.'))

#package_dir = {name: name}

# Data files used e.g. in tests
#package_data = {name: [os.path.join(name, 'tests', 'prt.txt')]}

# The current version number - MSI accepts only version X.X.X
#exec(open(os.path.join(name, 'version.py')).read())

# Scripts
scripts = []
for script_dir in script_directories:
    for dirname, dirnames, filenames in os.walk(script_dir):
        for filename in filenames:
            if not filename.endswith('.bat'):
                scripts.append(os.path.join(dirname, filename))

# Provide bat executables in the tarball (always for Win)
# if 'sdist' in sys.argv or os.name in ['ce', 'nt']:
#     for s in scripts[:]:
#         scripts.append(s + '.bat')

# Data_files (e.g. doc) needs (directory, files-in-this-directory) tuples

data_files = []

for data_dir, file_ext, tdir in data_directories:
    for dirname, dirnames, filenames in os.walk( os.path.join(data_dir)):
        fileslist = []
        for filename in filenames:
            _, ext = os.path.splitext(filename)
            if (not file_ext) or (ext == file_ext):
                fullname = os.path.join(dirname, filename)
                fileslist.append(fullname)
                #print(('share/' + dirname, fileslist))
        if tdir:
            data_files.append(( os.path.join('share' , tdir), fileslist ))
        else:
            data_files.append(( os.path.join('share' , dirname), fileslist ))

print("---------------")
print(data_files)
print("---------------")

#####

for path in ["README.md", "readme.md", "readme.txt"]:
    try:
        readme = open('README.md').read()
    except Exception as er:
        print("No readme file : ", er)
    else:
        break
else:
    readme = ""


setup(
    name=name,
    version=version,
    author=author,
    author_email=author_email,
    packages=packages,
    scripts=scripts,
    data_files=data_files,
    license=license,
    long_description=readme,
    install_requires=install_requires,
    dependency_links=dependency_links,
    long_description_content_type='text/markdown'
)



MAPFILE = "mapfile"
INTERFACE = "insterface"
IDENTIFIER = "identifier"

PREFIS = "prefix"
SIMULATED =  "simulated"
IGNORED = "ignored"
ADDRESS = "address"
SIMADDR = "simaddr"


PREFIX = "prefix"
FITS_PREFIX = "fits_prefix"

CTRL_CONFIG = "ctrl_config"
INITIALISATION = "initialisation"

SERVER_ID = 'server_id'
DEVICES = 'devices'
CFGFILE = 'cfgfile'
NAMESPACE = "namespace"

STATE = "state"
SUBSTATE = "substate"
STATUS = "status"

CFG = "cfg"
STAT = "stat"
RPC = "rpc"
TYPE = "type"



import os
__module_dir__, _ = os.path.split(__file__)
#__module_dir__, _ = os.path.split(__module_dir__) # so far client and server are in the same package
__package_dir__, __pkg__ = os.path.split(__module_dir__)
uis_path =[]
uis_path.append(os.path.join(__package_dir__, "share", "pydevmgr_qt_uis"))

def find_ui(name):
    for dirname in uis_path:
        path = os.path.join(dirname, name)
        if os.path.exists(path):
            return path
    raise ValueError('could not found ui file  %r in any of %r'%(name, uis_path))



from PyQt5 import QtWidgets, uic
from .io import find_ui
from .generic import record_widget, method_caller, method_switcher, RpcError
from pydevmgr import Adc, Prefixed
from .basic import BaseCtrl, BaseLine


class MOVE_MODE:
    ABSOLUTE = 0
    RELATIVE = 1
    VELOCITY = 2
    txt = {
       ABSOLUTE : "ABSOLUTE", 
       RELATIVE : "RELATIVE", 
       VELOCITY : "VELOCITY"
    }
    count = 3

class AXES:
    BOTH = 0
    AXIS1 = 1
    AXIS2 = 2
    txt = {
        BOTH  : 'BOTH AXES', 
        AXIS1 : 'AXIS 1',
        AXIS2 : 'AXIS 2'
    }
    count = 3


class AdcCtrl(BaseCtrl):
    
    def initUI(self):
        uic.loadUi(find_ui('adc_ctrl_frame.ui'), self)
        
    def update_device(self, data, adc):
        super(AdcCtrl, self).update_device(data, adc)
        #self.input_pos_target1.update()
        #self.input_pos_target2.update()
        #self.input_velocity.update() 
        
        if self.move_mode.currentIndex()==MOVE_MODE.VELOCITY: # VELOCITY
            self.input_pos_target.setEnabled(False)
        else:
            self.input_pos_target.setEnabled(True)
        
        
        self.motor1_pos_actual.setText( "{:.3f}".format(data['motor1.pos_actual']) )
        self.motor2_pos_actual.setText( "{:.3f}".format(data['motor2.pos_actual']) )
        self.motor1_pos_error.setText( "{:.3E}".format(data['motor1.pos_error']) )
        self.motor2_pos_error.setText( "{:.3E}".format(data['motor2.pos_error']) )
        
        self.motor1_pos_target.setText( "{:.3f}".format(data['motor1.pos_target']) )
        self.motor2_pos_target.setText( "{:.3f}".format(data['motor2.pos_target']) )
        self.motor1_vel_actual.setText( "{:.3f}".format(data['motor1.vel_actual']) )
        self.motor2_vel_actual.setText( "{:.3f}".format(data['motor2.vel_actual']) )
        
        self.track_mode_txt.setText( "{}".format(data['track_mode_txt']) )
        
    def link_device(self, downloader, adc, altname=None):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        adc (:class:`pydevmgr.Adc`):  Adc device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        super(AdcCtrl, self).link_device(downloader, adc, altname)
        token = self.token # comming from super 
        try:
            self.move.clicked.disconnect()
            self.start_track.disconnect()
            self.stop_track.disconnect()
        except TypeError:
            pass
        
        # init some field to what the state is 
        
        if adc.is_connected():
            self.input_velocity.setText("%.3f"%adc.motor1.cfg.velocity.get())
            self.input_pos_target.setText("%.3f"%adc.motor1.stat.pos_actual.get())   
        else:
            # put some dummy values
            self.input_velocity.setText("1.00")
            self.input_pos_target.setText("0.00")   
        self.input_angle.setText("0.0")
        
        self.input_axis.clear()
        self.input_axis.addItems(AXES.txt[i] for i in range(AXES.count))    
        
        self.move_mode.clear()
        self.move_mode.addItems(MOVE_MODE.txt[i] for i in range(MOVE_MODE.count))
        
        # add necessary nodes to the downloader state, substate, status, error_code 
        # are handled  by BaseCtrl.link_device
        
        downloader.add_node(token, 
                         adc.stat.track_mode_txt, 
                         adc.motor1.stat.pos_actual, adc.motor1.stat.pos_error, 
                         adc.motor1.stat.pos_target, adc.motor1.stat.vel_actual,# stat.vel_target,
                         adc.motor2.stat.pos_actual, adc.motor2.stat.pos_error, 
                         adc.motor2.stat.pos_target, adc.motor2.stat.vel_actual,# stat.vel_target,
                        )
        
        
        def move(idx, axis, pos, vel):
            if idx == MOVE_MODE.ABSOLUTE:
                adc.move_abs(axis, pos, vel)
            elif idx == MOVE_MODE.RELATIVE:
                adc.move_rel(axis, pos, vel)
            elif idx == MOVE_MODE.VELOCITY:
                adc.move_vel(axis, vel)
            
        # Now link the buttons to an action
        move = method_caller(move, 
                            [self.move_mode.currentIndex, 
                             self.input_axis.currentIndex,
                             self.input_pos_target.text, 
                             self.input_velocity.text
                             ],
                             self.feedback
                            )
        self.move.clicked.connect(move)
        
        stop = method_caller(adc.stop, [], self.feedback)
        self.stop.clicked.connect(stop)
        
        
        start_track = method_caller(adc.start_track, [self.input_angle.text], self.feedback)
        self.start_track.clicked.connect(start_track)
        
        stop_track = method_caller(adc.stop_track, [], self.feedback)
        self.stop_track.clicked.connect(stop_track)
        
        move_angle = method_caller(adc.move_angle,
                                   [self.input_angle.text], 
                                    self.feedback
                                   )
        self.move_angle.clicked.connect(move_angle)

class AdcLine(BaseLine):
    
    def initUI(self):
        uic.loadUi(find_ui('adc_line_frame.ui'), self)
        
    def update_device(self, data, adc):
        super(AdcLine, self).update_device(data, adc)
        
        
        self.motor1_pos_actual.setText( "{:.3f}".format(data['motor1.pos_actual']) )
        self.motor2_pos_actual.setText( "{:.3f}".format(data['motor2.pos_actual']) )
        
        self.track_mode_txt.setText( "{}".format(data['track_mode_txt']) )
    
    
    def link_device(self, downloader, adc, altname=None, more_methods=[]):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        adc (:class:`pydevmgr.Adc`):  Adc device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        methods = [
             "---", # is a separator and is not counted as item 
            ("MOVE ANGLE", adc.move_angle, [self.input_angle.text]),
            ("START TRACK", adc.start_track,[self.input_angle.text]), 
            ("STOP TRACK", adc.stop_track, [] ), 
        ]+ list(more_methods)
                        
        super(AdcLine, self).link_device(downloader, adc, altname, more_methods=methods)
        token = self.token # comming from super 
                
        self.input_angle.setText("0.00")            
        
        downloader.add_node(token, 
                         adc.motor1.stat.pos_actual, adc.motor2.stat.pos_actual, 
                         adc.stat.track_mode_txt
                        )
        

record_widget("ctrl", "Adc", AdcCtrl)
record_widget("line", "Adc", AdcLine)




from .device import Device, _name_attr, _inc, GROUP
from . import keys as K
from .manager import Manager # see at the end of file
from .interface import nodealiasproperty, Int16, Int32
from collections import OrderedDict 
from .tools import upload
#                      _              _   
#   ___ ___  _ __  ___| |_ __ _ _ __ | |_ 
#  / __/ _ \| '_ \/ __| __/ _` | '_ \| __|
# | (_| (_) | | | \__ \ || (_| | | | | |_ 
#  \___\___/|_| |_|___/\__\__,_|_| |_|\__|
# 

##### ############
# Sequence
@_name_attr
class INITSEQ:
    END = 0
    FIND_INDEX = 1
    FIND_REF_LE = 2
    FIND_REF_UE = 3
    FIND_LHW = 4
    FIND_UHW = 5

    DELAY = 6
    MOVE_ABS = 7
    MOVE_REL = 8
    CALIB_ABS = 9
    CALIB_REL = 10
    CALIB_SWITCH = 11

##### ############
# SUBSTATE
@_name_attr
class SUBSTATE:
    NONE =  0

    NOTOP_NOTREADY =  100
    NOTOP_READY = 101
    NOTOP_INITIALIZING = 102
    NOTOP_ABORTING = 107
    NOTOP_CLEARING_NOVRAM = 108

    NOTOP_ERROR =  199

    OP_STANDSTILL =216
    OP_MOVING = 217
    OP_SETTING_POS = 218
    OP_STOPPING = 219

    OP_ERROR =299
    group = {
        NONE                   : GROUP.UNKNOWN,
        NOTOP_NOTREADY         : GROUP.NOK,
        NOTOP_READY            : GROUP.NOK,
        NOTOP_INITIALIZING     : GROUP.BUZY,
        NOTOP_ABORTING         : GROUP.BUZY,
        NOTOP_CLEARING_NOVRAM  : GROUP.BUZY,
        NOTOP_ERROR            : GROUP.ERROR, 
        OP_STANDSTILL          : GROUP.OK, 
        OP_MOVING              : GROUP.BUZY, 
        OP_SETTING_POS         : GROUP.BUZY,
        OP_STOPPING            : GROUP.BUZY,
        OP_ERROR               : GROUP.ERROR,    
    }


### ############
# Motor ERROR
@_name_attr
class ERROR:
    OK	                   = _inc(0)
    HW_NOT_OP              = _inc()
    LOCAL                  = _inc()
    INIT_ABORTED           = _inc()
    TIMEOUT_INIT           = _inc()
    TIMEOUT_MOVE           = _inc()
    TIMEOUT_RESET          = _inc()
    TIMEOUT_SETPOS         = _inc()
    TIMEOUT_USER_PREINIT   = _inc()
    TIMEOUT_USER_POSTINIT  = _inc()
    TIMEOUT_USER_PREMOVE   = _inc()
    TIMEOUT_USER_POSTMOVE  = _inc()
    SETPOS                 = _inc()
    STOP                   = _inc()
    ABORT                  = _inc()
    SW_LIMIT_LOWER         = _inc()
    SW_LIMIT_UPPER         = _inc()
    BRAKE_ACTIVE           = _inc()
    BRAKE_ENGAGE           = _inc()
    BRAKE_DISENGAGE        = _inc()
    SWITCH_NOT_USED        = _inc()
    ENABLE                 = _inc()
    NOVRAM_READ            = _inc()
    NOVRAM_WRITE           = _inc()
    SWITCH_EXIT            = _inc()
    STOP_LIMITS_BOTH       = _inc()
    HW_LIMITS_BOTH         = _inc()
    IN_POS                 = _inc()
    LOCKED                 = _inc()
    SoE_ADS_ERROR          = _inc()
    SoE_SERCOS_ERROR       = _inc()

    # Simulator errors
    SIM_NOT_INITIALISED			= 90
    SIM_NULL_POINTER			= 100	

    # TwinCAT errors
    TC_VEL						= 16929	
    TC_NOT_READY_FOR_START		= 16933	
    TC_DISABLED_MOVE			= 16992	
    TC_BISECTION				= 17022	
    TC_MODULO_POS				= 17026	
    TC_STOP_ACTIVE				= 17135	
    TC_VEL_NEG					= 17241	
    TC_TARGET_LSW				= 17504	
    TC_TARGET_USW				= 17505	
    TC_FOLLOWING_ERROR			= 17744	
    TC_NOT_READY				= 18000	
    TC_IN_POS_6_SEC				= 19207	
    txt = {
    	OK:					  'OK',
    	HW_NOT_OP:			  'ERROR: TwinCAT not OP or CouplerState not mapped.',
    	LOCAL:				  'ERROR: Control not allowed. Motor in Local mode.',
    	INIT_ABORTED:		  'ERROR: INIT command aborted.',
    	TIMEOUT_INIT:		  'ERROR: INIT timed out.',
    	TIMEOUT_MOVE:		  'ERROR: Move timed out.',
    	TIMEOUT_RESET:		  'ERROR: Reset timed out.',
    	TIMEOUT_SETPOS:		  'ERROR: Set Position timed out.',
    	TIMEOUT_USER_PREINIT: 'ERROR: User PRE-INIT timed out.',
    	TIMEOUT_USER_POSTINIT:'ERROR: User POST-INIT timed out.',
    	TIMEOUT_USER_PREMOVE: 'ERROR: User PRE-MOVE timed out.',
    	TIMEOUT_USER_POSTMOVE:'ERROR: User POST-MOVE timed out.',
    	SETPOS:				  'ERROR: Set Position failed.',
    	STOP:				  'ERROR: STOP failed.',
    	
    	ABORT:				  'ERROR: Motion aborted.',
    	SW_LIMIT_LOWER:		  'ERROR: Lower SW Limit Exceeded.',
    	SW_LIMIT_UPPER:		  'ERROR: Upper SW Limit Exceeded.',
    	BRAKE_ACTIVE:		  'ERROR: Cannot move. Brake active.',
    	BRAKE_ENGAGE:		  'ERROR: Failed to engage brake.',
    	BRAKE_DISENGAGE:	  'ERROR: Failed to disengage brake.',
    	SWITCH_NOT_USED:	  'ERROR: Switch was not detected in previous INIT action.',
    	ENABLE:				  'ERROR: Failed to enable Axis.',
    	NOVRAM_READ:		  'ERROR: Failed to read from NOVRAM',
    	NOVRAM_WRITE:		  'ERROR: Failed to write to NOVRAM',
    	SWITCH_EXIT:		  'ERROR: Timeout on switch exit. Check nTimeoutSwitch.',
    	STOP_LIMITS_BOTH:	  'ERROR: Both LSTOP and USTOP limits active.',
    	HW_LIMITS_BOTH:		  'ERROR: Both limit switches LHW and UHW active.',
    	IN_POS:				  'ERROR: In-Pos switch not active at the end of movement.',
    	LOCKED:				  'ERROR: Motor Locked! Cannot move.',
    	SoE_ADS_ERROR:		  'ERROR: SoE ADS Error.',
    	SoE_SERCOS_ERROR:	  'ERROR: SoE Sercos Error.',

    	SIM_NOT_INITIALISED:  'ERROR: Simulator not initialised.',
    	SIM_NULL_POINTER:	  'ERROR: Simulator input parameter is a NULL pointer.',
    	
    	# Beckhoff TwinCAT most common errors
    	TC_VEL:				    'ERROR: Requested set velocity is not allowed.',
    	TC_NOT_READY_FOR_START: 'ERROR: Drive not ready during axis start. Maybe SW limits.',
    	TC_DISABLED_MOVE:		'ERROR: Motor disabled while moving. Reset required!',
    	TC_BISECTION:			'WARNING: Motion command could not be realized (BISECTION)',
    	TC_MODULO_POS:		  'ERROR: Target position >= full turn (modulo-period)',
    	TC_STOP_ACTIVE:		  'ERROR: Stop command still active. Axis locked. Reset required!',
    	TC_VEL_NEG:			  'ERROR: Set velocity not allowed (<=0)',
    	TC_TARGET_LSW:		  'ERROR: Target position beyond Lower Software Limit.',
    	TC_TARGET_USW:		  'ERROR: Target position beyond Upper Software Limit.',
    	TC_FOLLOWING_ERROR:	  'ERROR: Following error. Reset required!',
    	TC_NOT_READY:		  'ERROR: Drive not ready for operation.',
    	TC_IN_POS_6_SEC:	  'ERROR: In-position 6 sec timeout. Reset required!',
    }
# e.g. ERROR.txt.get( er_code, "Unknown Error") -> return the error text



### ##############
# RPC error
@_name_attr
class RPC_ERROR:
    OK =  0
    NOT_OP =  -1
    NOT_NOTOP_READY =  -2
    NOT_NOTOP_NOTREADY = -3
    LOCAL =  -4
    SW_LIMIT_LOWER = -5
    SW_LIMIT_UPPER = -6
    INIT_WHILE_MOVING = -7
    txt = {
        OK:					 'OK',
        NOT_OP:				 'Cannot control motor. Not in OP state.',
        NOT_NOTOP_READY:	 'Call failed. Not in NOTOP_READY.',
        NOT_NOTOP_NOTREADY:	 'Call failed. Not in NOTOP_NOTREADY/ERROR.',
        LOCAL:				 'RPC calls not allowed in Local mode.',
        SW_LIMIT_LOWER:		 'Move rejected. Target Pos < Lower SW Limit',
        SW_LIMIT_UPPER:		 'Move rejected. Target Pos > Upper SW Limit',
        INIT_WHILE_MOVING:	 'Cannot INIT moving motor. Motor stopped. Retry.',
    }
    
@_name_attr
class AXIS_TYPE:
    LINEAR= 1
    CIRCULAR=2
    CIRCULAR_OPTIMISED = 3

def axis_type(value):
    """ return always a axis_type int number from a number or a string
    
    Raise a ValueError if the input string does not match axis type
    Example:
        axis_type('LINEAR') == 1
        axis_type(1) == 1
    """
    try:
        val = getattr(AXIS_TYPE, value) if isinstance(value, str) else value
    except AttributeError:
        raise ValueError('Unknown AXIS type %r'%value)
    return Int32(val)
# 
#   __                  _   _                 
#  / _|_   _ _ __   ___| |_(_) ___  _ __  ___ 
# | |_| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# |  _| |_| | | | | (__| |_| | (_) | | | \__ \
# |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
# 

def init_sequence_to_cfg(initialisation, INITSEQ=INITSEQ):
    """ from a config initialisation dict return a dictionary of key/value for .cfg interface """            
    # set the init sequence
    
    cfg_dict = {} 
        
    sequence = initialisation['sequence']
    
    # reset all sequence variable
    for i in range(1,11):
        cfg_dict["init_seq{}_action".format(i)] = INITSEQ.END
        cfg_dict["init_seq{}_value1".format(i)] = 0.0
        cfg_dict["init_seq{}_value2".format(i)] = 0.0
        
    # in mapping file the sequence number start from 1
    for seqnum, action_name in enumerate(sequence, start=1):
        try:
            action_number = getattr(INITSEQ, action_name)
        except AttributeError:
            raise ValueError('Unknown init sequence action %r'%action_name)

        try:
            action_def = initialisation[action_name]
        except KeyError:
            raise ValueError("Definition of init sequence action %r is missing"%action_name)

        try:
            value1 = action_def['value1']
        except KeyError:
            raise ValueError("Could not find 'value1' in inisequence at action %r"%action_name)

        try:
            value2 = action_def['value2']
        except KeyError:
            raise ValueError("Could not find 'value2' in inisequence at action %r"%action_name)

        cfg_dict["init_seq%d_action"%seqnum] = action_number
        cfg_dict["init_seq%d_value1"%seqnum] = value1
        cfg_dict["init_seq%d_value2"%seqnum] = value2        
    return cfg_dict




#  _       _             __                
# (_)_ __ | |_ ___ _ __ / _| __ _  ___ ___ 
# | | '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \
# | | | | | ||  __/ |  |  _| (_| | (_|  __/
# |_|_| |_|\__\___|_|  |_|  \__,_|\___\___|

class MotorStatInterface(Device.StatInterface): 
    ERROR = ERROR
    SUBSTATE = SUBSTATE
    
    @nodealiasproperty("is_moving", ["substate"])
    def is_moving(self, substate):
        """ -> True is axis is moving """
        return substate == self.SUBSTATE.OP_MOVING

    @nodealiasproperty("is_standstill", ["substate"])
    def is_standstill(self,  substate):
        """ -> True is axis is standstill """
        return substate == self.SUBSTATE.OP_STANDSTILL
    
    _mot_positions = None# will be overwriten by motor 
    @nodealiasproperty("pos_name", ["pos_actual"])
    def pos_name(self, pos_actual):
        if not self._mot_positions: return ''
        positions = self._mot_positions
        tol = positions['tolerance']
        for pname in positions['posnames']:
            if abs(positions[pname]-pos_actual)<tol:
                return pname
        return ''
    
class MotorCfgInterface(Device.CfgInterface):
    # we can define the type to parse value directly on the class by annotation
    axis_type : axis_type
    tout_init: Int32 
    tout_move: Int32        
    tout_switch: Int32
    init_seq1_action: Int32
    init_seq2_action: Int32
    init_seq3_action: Int32
    init_seq4_action: Int32
    init_seq5_action: Int32
    init_seq6_action: Int32
    init_seq7_action: Int32
    init_seq8_action: Int32
    init_seq9_action: Int32
    init_seq10_action: Int32

# redefine the Method interface to include the proper description of the RpcError
class MotorRpcNode(Device.RpcInterface.RpcNode):
    RPC_ERROR = RPC_ERROR

class MotorRpcInterface(Device.RpcInterface):
    RpcNode = MotorRpcNode
    RPC_ERROR = RPC_ERROR
    
    
    ##
    # the type of rpcMethod argument can be defined by annotation
    # All args types must be defined in a tuple
    rpcMoveAbs : (float, float)
    rpcMoveRel : (float, float)
    rpcMoveVel : (float,)

#      _            _          
#   __| | _____   _(_) ___ ___ 
#  / _` |/ _ \ \ / / |/ __/ _ \
# | (_| |  __/\ V /| | (_|  __/
#  \__,_|\___| \_/ |_|\___\___|
#
class Motor(Device):
    SUBSTATE = SUBSTATE
    ERROR = ERROR
    
    INITSEQ = INITSEQ
    AXIS_TYPE = AXIS_TYPE

    RpcInterface = MotorRpcInterface
    StatInterface = MotorStatInterface
    CfgInterface = MotorCfgInterface
    
    def get_configuration(self, **kwargs):
        """  return a node/value pair dictionary ready to be uploaded 
        
        The node/value dictionary represent the device configuration. 
        
        Args:
            **kwargs : name/value pairs pointing to cfg.name node
                      This allow to change configuration on the fly
                      without changing the config file. 
        """
        
        config = self._config 
        
        ctrl_config = config.get(K.CTRL_CONFIG, {})  
        # just update what is in ctrl_config, this should work for motor 
        # one may need to check parse some variable more carefully       
        cfg_dict = {self.cfg.get_node(k):v for k,v in  ctrl_config.items() }
        cfg_dict.update({self.cfg.get_node(k):v for k,v in  kwargs.items() })
        
        
        try:
            initialisation = config[K.INITIALISATION]
        except KeyError:
            pass
        else:
            init_cfg = init_sequence_to_cfg(initialisation, self.INITSEQ)
            cfg_dict.update({self.cfg.get_node(k):v for k,v in init_cfg.items()})
        
        ###
        # Set the new config value to the device 
        return cfg_dict
          
    
    ### redefine stat property to include positions 
    @property
    def stat(self):
        if self._stat is None:
            # TODO add the parser functions dictionay
            self._stat = self.StatInterface(self.key, self._ua_client, self._map[K.STAT], self.ua_prefix, ua_namespace=self.ua_namespace )
            self._stat._mot_positions = self._config.get('positions', None)
        return self._stat
    
    @property
    def posnames(self):
        """ configured position names in a name:(pos, tol) dictionary """
        try:
            return self.__dict__['posnames']
        except KeyError:                             
            try:
                positions = self._config['positions']
                tol = positions['tolerance']
                posnames = OrderedDict( (name,(positions[name], tol)) for name in positions['posnames'])
            except KeyError as e:
                posnames = {}
            # cash the posnames
            self.__dict__['posnames'] =  posnames    
            return posnames          
        
    
    def clear(self):
        """ Clear cashed values """
        super(Motor, self).clear()
        self.__dict__.pop('posnames', None)
    
    def move_abs(self, absPos, vel):
        """ move motor to an absolute position 
        
        self.move_abs(pos, vel) <-> self.rpc.rpcMoveAbs(pos, vel)
        
        Args:
            absPos (float): absolute position
            vel (float):   target velocity for the movement
            
        """
        self.rpc.rpcMoveAbs.rcall(absPos, vel)
    
    def move_name(self, name, vel):
        """ move motor to a named position 
        
        Args:
           name (str): named position
           vel (float):   target velocity for the movement
        """
        absPos = self.get_pos_target_of_name(name)
        self.rpc.rpcMoveAbs.rcall(absPos, vel)
        
    def move_rel(self, relPos, vel):
        """ Move motor relative position
        
        Args:
           relPos (float): relative position
           vel (float):   target velocity for the movement
        """
        self.rpc.rpcMoveRel.rcall(relPos, vel)

    def move_vel(self, vel):
        """ Move motor in velocity mode 
        
        Args:
           vel (float): target velocity
        """
        self.rpc.rpcMoveVel.rcall(vel)

    def stop(self):
        """ Stop the motor """
        self.rpc.rpcStop.rcall()

    
    def get_pos_target_of_name(self, name):
        """return the configured target position of a given pos name or raise error"""
        try:
            position = self._config['positions'][name]
        except KeyError:
            raise ValueError('unknown posname %r'%name)
        return position

    def get_name_of_pos(self, pos_actual):
        """ Retrun the name of a position from a position as input or ''
        
        Example:
            m.get_name_of( m.stat.pos_actual.get() )
        """
        try:
            positions = self._config['positions']
        except KeyError:
            return ''

        tol = positions['tolerance']
        for pname in positions['posnames']:
            if abs(positions[pname]-pos_actual)<tol:
                return pname
        return ''
        
    def is_near(self, pos, tol, data=None):
        """ -> True when abs(pos_actual-pos)<tol """
        apos = self.stat.pos_actual.get(data) 
        return abs(apos-pos)<tol
    
Manager.record_new_device_type('Motor', Motor)

Manager.record_cfg_template('Motor', """{name}:
  type: Motor
  interface: Softing
  identifier: PLC1
  prefix: MAIN.Motor1 # TO BE UPDATED  
  simulated: false
  ignored: false
  address: {address}   # TO BE UPDATED 
  simaddr: opc.tcp://127.0.0.1:7578
  mapfile: "{cfg_dir}/mapMotor.yml"  # TO BE UPDATED 
  fits_prefix: "MOT1"
  ctrl_config:
    velocity:              3.0
    min_pos:               0.0
    max_pos:               359.0
    axis_type:             CIRCULAR
    active_low_lstop:      false
    active_low_lhw:        false
    active_low_ref:        true
    active_low_index:      false
    active_low_uhw:        true
    active_low_ustop:      false
    brake:                 false
    low_brake:             false
    low_inpos:             false
    backlash:              0.0
    tout_init:             30000
    tout_move:             120000
    tout_switch:           10000
  initialisation:
      sequence: ['FIND_LHW', 'FIND_UHW', 'CALIB_ABS', 'END']
      FIND_LHW:
         value1: 4.0
         value2: 4.0
      FIND_UHW:
         value1: 4.0
         value2: 4.0
      CALIB_ABS:
         value1: 0.0
         value2: 0.0
      END:
         value1: 0.0
         value2: 0.0
  positions:
     posnames: ['ON', 'OFF']
     tolerance: 1                              # Tolerance in UU
     "ON": 30
     "OFF": 100
""")


